import datetime

from .conditions import Condition, CompareCondition


NAME = 'Examples'


class FullName(Condition):

    condstr = 'FULL_NAME'

    def eval_bool(self, user, **kwargs):
        return bool(user.first_name and user.last_name)


class Level(CompareCondition):

    condstr = 'LEVEL'

    def eval_operand(self, user, **kwargs):
        return user.userprofile.level


class DateJoined(CompareCondition):

    condstr = 'DATE_JOINED'
    cast_operand = lambda self, timestamp: datetime.datetime.strptime(timestamp, "%m/%d/%Y")

    @classmethod
    def operators(cls):
        return {
            '<': datetime.datetime.__lt__,
            '<=': datetime.datetime.__le__,
            '==': datetime.datetime.__eq__,
            '!=': datetime.datetime.__ne__,
            '>=': datetime.datetime.__ge__,
            '>': datetime.datetime.__gt__,
        }

    def eval_operand(self, user, **kwargs):
        return user.date_joined.strftime("%m/%d/%Y")


class EmailDomain(Condition):

    condstr = 'EMAIL_DOMAIN'

    def eval_bool(self, user, **kwargs):
        return user.email.split('@')[1] == self.key



import datetime

from django.contrib.auth.models import User
from django.db import models
from django.test import TestCase

from . import examples as condition_examples
from .fields import ConditionsField
from .lists import eval_conditions
from .types import conditions_from_module


class UserProfile(models.Model):

    user = models.OneToOneField(User)
    level = models.IntegerField(default=1)


class Campaign(models.Model):

    text = models.TextField()
    conditions = ConditionsField(definitions=conditions_from_module(condition_examples))


class CampaignTest(TestCase):

    def setUp(self):
        self.mrx = User.objects.create_user(
            username='mr.x',
            email='x@gmail.com',
            password='top_secret'
        )
        self.mrsy = User.objects.create_user(
            username='mrs.y',
            email='y@yahoo.com',
            password='also_top_secret'
        )
        # Create UserProfile objects
        for user in [self.mrx, self.mrsy]:
            UserProfile.objects.create(user=user)

        # Set Mr. X to have been a long-term member
        self.mrx.date_joined = datetime.datetime(day=31, month=12, year=2013)

        self.campaign = Campaign.objects.create(
            text="Thanks for providing your full name.",
            conditions={
                'all': ["FULL_NAME",],
            }
        )
        self.comparecondition_campaign = Campaign.objects.create(
            text="Congratulations on getting to Level 5!",
            conditions={
                'all': ["LEVEL == 5",],
            }
        )
        self.long_term_user_campaign = Campaign.objects.create(
            text="Thanks for being a long-term member.",
            conditions={
                'any': ["DATE_JOINED < 01/01/2014",],
            }
        )
        self.non_gmail_users_campaign = Campaign.objects.create(
            text="Why aren\'t you using Gmail?",
            conditions={
                'all': ["NOT EMAIL_DOMAIN gmail.com"]
            }
        )
        self.long_term_gmail_yahoo_user_campaign = Campaign.objects.create(
            text="You\'ve been using the same email for a long time.",
            conditions={
                'all': [
                    {
                        'any': ["EMAIL_DOMAIN gmail.com", "EMAIL_DOMAIN yahoo.com",],
                    },
                    "DATE_JOINED < 01/01/2014",
                ],
            }
        )

    def tearDown(self):
        Campaign.objects.all().delete()
        UserProfile.objects.all().delete()
        User.objects.all().delete()

    def test_basic_campaign_targetting(self):
        # Neither user should be targetted before having a name
        self.assertFalse(eval_conditions(self.campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.campaign, 'conditions', self.mrsy))

        # Mr. X adds his full name; Mrs. Y only adds her last name
        self.mrx.first_name = 'Mr.'
        self.mrx.last_name = 'X'
        self.mrsy.last_name = 'Y'

        # Now Mr. X is targetted, but Mrs. Y is not
        self.assertTrue(eval_conditions(self.campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.campaign, 'conditions', self.mrsy))

    def test_comparecondition_campaign(self):
        # Neither user should be targetted with level 1
        self.assertFalse(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrsy))

        # Mrs. Y upgrades to level 5
        self.mrsy.userprofile.level += 4

        # Now Mrs. Y is targetted, but Mr. X is not
        self.assertFalse(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrx))
        self.assertTrue(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrsy))

        # Mrs. Y upgrades to level 6
        self.mrsy.userprofile.level += 1

        # Now neither user once again is targetted
        self.assertFalse(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.comparecondition_campaign, 'conditions', self.mrsy))

    def test_custom_operators(self):
        # Mr. X is a long-term member, but Mrs. Y isn't
        self.assertTrue(eval_conditions(self.long_term_user_campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.long_term_user_campaign, 'conditions', self.mrsy))

    def test_key(self):
        # Mr X. uses Gmail, but Mrs. Y doesn't
        self.assertFalse(eval_conditions(self.non_gmail_users_campaign, 'conditions', self.mrx))
        self.assertTrue(eval_conditions(self.non_gmail_users_campaign, 'conditions', self.mrsy))

    def test_nested_condlists(self):
        # Mr. X is a long-term member, but Mrs. Y isn't
        # Even though they both use either Gmail or Yahoo!, only Mr. X is targeted
        self.assertTrue(eval_conditions(self.long_term_gmail_yahoo_user_campaign, 'conditions', self.mrx))
        self.assertFalse(eval_conditions(self.long_term_gmail_yahoo_user_campaign, 'conditions', self.mrsy))



"""
:Created: 31 October 2014
:Author: Lucas Connors

"""

import abc
import inspect
import logging
import random

from .exceptions import UndefinedConditionError, InvalidConditionError


__all__ = ['Condition', 'CompareCondition',]


logger = logging.getLogger('condition')


class Condition(object):
    """
    A condition which evaluates either to True or False
    """

    __metaclass__ = abc.ABCMeta
    keys_allowed = [] # By default, allow any key
    key_examples = []
    operand_examples = []

    def __repr__(self):
        return "{not_}{condstr}{key}{operator}".format(
            not_='NOT ' if self.include_not else '',
            condstr=self.condstr,
            key=' {key}'.format(key=self.key) if self.key else '',
            operator=' {operator} {operand}'.format(
                operator=self.operatorname,
                operand=self.operand
            ) if self.operatorname else ''
        )

    encode = __repr__

    @classmethod
    def decode(cls, value, definitions=None):
        all_conditions = definitions or {}
        condition_definitions = {}
        for condition_group in all_conditions.values():
            condition_definitions.update(condition_group)
        include_not = False
        operator = operand = key = None
        condtuple = value.split(' ')

        # CONDSTR
        condstr = condtuple.pop(0)

        # NOT
        if condstr == 'NOT':
            include_not = True
            condstr = condtuple.pop(0)

        # Evaluate CONDSTR
        try:
            condition = condition_definitions[condstr]
        except KeyError:
            raise UndefinedConditionError("The condition {condstr} is not defined.".format(condstr=condstr))

        # KEY
        if len(condtuple) > 0:
            key = condtuple.pop(0)
            if issubclass(condition, CompareCondition) and key in condition.operators():
                operator = key
                key = None
                if condition.key_required():
                    raise InvalidConditionError("A key is required in this condition.")
            elif condition.key_required() and condition.keys_allowed and key not in condition.keys_allowed:
                raise InvalidConditionError(
                    "The key in this condition must be one of: {keys}; found {key}".format(
                        keys=', '.join(condition.keys_allowed),
                        key=key
                    )
                )

        # OPERATOR
        if not operator and len(condtuple) > 0:
            operator = condtuple.pop(0)

        # OPERAND
        if operator:
            operand = condtuple.pop(0)

        # Return initialized condition
        return condition(operator=operator, operand=operand, key=key, include_not=include_not)

    @classmethod
    def key_required(cls):
        try:
            eval_func = cls.eval_operand
        except AttributeError:
            eval_func = cls.eval_bool

        try:
            func_code = eval_func.__code__
        except AttributeError: # Python 2
            func_code = eval_func.im_func.func_code

        return 'key' in func_code.co_names

    @classmethod
    def key_example(cls):
        if cls.key_required():
            key = ' '
            if cls.keys_allowed:
                key += random.choice(cls.keys_allowed)
            elif cls.key_examples:
                key += random.choice(cls.key_examples)
            else:
                key += 'SOME_KEY_HERE'
        else:
            key = ''
        return key

    @classmethod
    def module_name(cls):
        module = inspect.getmodule(cls)
        try:
            return module.NAME
        except AttributeError:
            return module.__name__.split('.')[-1]

    @classmethod
    def help_text(cls):
        """
        Randomly generate a possible example for this condition
        """
        return "Ex. {not_}{condstr}{key}".format(
            not_=random.choice(['NOT ', '']),
            condstr=cls.condstr,
            key=cls.key_example()
        )

    @classmethod
    def full_description(cls):
        """
        Return full description of condition (from docstring), if available
        (otherwise return help_text)
        """
        docstring = inspect.getdoc(cls)
        if docstring:
            return docstring.replace('\r', '').replace('\n', ' ')
        else:
            return cls.help_text()

    def __init__(self, operator=None, operand=None, key=None, include_not=False, *args, **kwargs):
        super(Condition, self).__init__(*args, **kwargs)
        self.operatorname = operator
        self.key = key
        self.include_not = include_not

    def eval(self, user, **kwargs):
        try:
            evaluation = self.eval_bool(user, **kwargs)
        except Exception as e:
            logger.warning("An exception occurred while processing a condition: {exception}".format(exception=str(e)), exc_info=True)
            return False

        if self.include_not:
            return not evaluation
        else:
            return evaluation

    @abc.abstractmethod
    def eval_bool(self, user, **kwargs):
        pass


class CompareCondition(Condition):
    """
    A condition which evaluates based on a comparison
    (ie. <, ==, >=, !=, etc.)
    """

    cast_operand = float

    @classmethod
    def operators(cls):
        """
        Produce the dict of operators for this condition
        """
        if cls.cast_operand in [float, int]:
            return {
                '<': lambda x, y: x < y,
                '<=': lambda x, y: x <= y,
                '==': lambda x, y: x == y,
                '!=': lambda x, y: x != y,
                '>=': lambda x, y: x >= y,
                '>': lambda x, y: x > y,
            }
        else:
            return {}

    @classmethod
    def operand_example(cls):
        if cls.cast_operand in [float, int]:
            if cls.cast_operand == float:
                operand = round(random.uniform(0, 100), 2)
            elif cls.cast_operand == int:
                operand = random.randint(0, 100)
        elif cls.operand_examples:
            operand = random.choice(cls.operand_examples)
        else:
            operand = 'SOME_OPERAND_HERE'
        return operand

    @classmethod
    def help_text(cls):
        """
        Randomly generate a possible example for this condition
        """
        normal_condition_help_text = super(CompareCondition, cls).help_text()

        help_text = "{normal} {operator} {operand}".format(
            normal=normal_condition_help_text,
            operator=random.choice(cls.operators().keys()),
            operand=cls.operand_example()
        )
        return help_text

    def __init__(self, operator, operand, key=None, include_not=False, *args, **kwargs):
        try:
            self.operator = self.operators()[operator]
        except KeyError:
            raise InvalidConditionError("The given function is not a comparison function or is unsupported.")
        self.operand = self.cast_operand(operand)

        super(CompareCondition, self).__init__(operator, operand, key, include_not, *args, **kwargs)

    def eval_bool(self, user, **kwargs):
        return self.operator(self.cast_operand(self.eval_operand(user, **kwargs)), self.operand)

    @abc.abstractmethod
    def eval_operand(self, user, **kwargs):
        pass
